// Copyright (C) 1999-2007
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <iostream>
#include <sstream>
#include <iomanip>
using namespace std;

#include <tcl.h>
#include <tk.h>
#include <Xlib.h>

#include <tkMacOSX.h>
#include <tkMacOSXInt.h>
#include <tkMacOSXFont.h>

#include "vector.h"

#define RGBFLOATRED(c)   ((c).red   / 65535.0)
#define RGBFLOATGREEN(c) ((c).green / 65535.0)
#define RGBFLOATBLUE(c)  ((c).blue  / 65535.0)

char* XImageData(XImage* ximage)
{
  GWorldPtr gw = TkMacOSXGetDrawablePort((Pixmap)(ximage->obdata));
  PixMapHandle ph = GetGWorldPixMap(gw);
  PixMap* pm = *ph;
  return GetPixBaseAddr(ph);
}

void XDrawRotString(Display* display, Drawable drawable, GC gc, Tk_Font font, 
		    double angle, const Vector& v, const char* text)
{
  // init
  unsigned long color = gc->foreground;
  MacDrawable* macWin = (MacDrawable*)drawable;

  GWorldPtr destPort = TkMacOSXGetDrawablePort(drawable);
  CGContextRef context;
  OSStatus err = QDBeginCGContext(destPort, &context);

  // set window clip rgn
  RgnHandle clipRgn = NewRgn();
  Rect bounds;
  GetPortBounds(destPort, &bounds);

  RectRgn(clipRgn, &bounds);
  SectRegionWithPortClipRegion(destPort, clipRgn);
  SectRegionWithPortVisibleRegion(destPort, clipRgn);

  // any other gc clipping rgns
  if (gc->clip_mask) {
    TkpClipMask* msk = (TkpClipMask*)gc->clip_mask;
    if (msk->type == TKP_CLIP_REGION) {
      int xOffset = macWin->xOff + gc->clip_x_origin;
      int yOffset = macWin->yOff + gc->clip_y_origin;
      RgnHandle rgn = NewRgn();
      CopyRgn((RgnHandle)msk->value.region, rgn);
      OffsetRgn(rgn, xOffset, yOffset);
      SectRgn(rgn, clipRgn, clipRgn);
      DisposeRgn(rgn);
    }
  }

  ClipCGContextToRegion(context, &bounds, clipRgn);
  DisposeRgn(clipRgn);
  SyncCGContextOriginWithPort(context, destPort);

  // rotate the text only
  CGAffineTransform mm = CGAffineTransformMakeRotation(angle);
  CGContextSetTextMatrix(context, mm);

  // color
  RGBColor macColor;
  if (TkSetMacColor(gc->foreground, &macColor) == true) {
    CGContextSetRGBFillColor(context,
			     RGBFLOATRED(macColor), 
			     RGBFLOATGREEN(macColor), 
			     RGBFLOATBLUE(macColor),
			     1);
    CGContextSetRGBStrokeColor(context,
			       RGBFLOATRED(macColor),
			       RGBFLOATGREEN(macColor),
			       RGBFLOATBLUE(macColor),
			       1);
  }

  // map to window space
  Vector center = 
    Vector((bounds.right-bounds.left)/2.,(bounds.bottom-bounds.top)/2.);
  Vector vv = v *
    Translate(-center) *
    Translate(macWin->xOff, macWin->yOff) *
    FlipY() *
    Translate(center);

  // draw it
  CGContextSetTextDrawingMode (context, kCGTextFill);

  TkFont* fontPtr = (TkFont*)font;
  Tcl_DString ds;
  Tcl_DStringInit(&ds);
  int size = Tk_PostscriptFontName(font, &ds);
  CGContextSelectFont(context, Tcl_DStringValue(&ds), 
		      size, kCGEncodingMacRoman);
  Tcl_DStringFree(&ds);

  CGContextShowTextAtPoint (context, vv[0], vv[1], text, strlen(text));

  /*
    ATSFontRef atsfont;
    if (fontPtr->fa.weight)
    atsfont = ATSFontFindFromPostScriptName(CFSTR("Times-Bold"), 
    kATSOptionFlagsDefault);
    else if (fontPtr->fa.slant)
    atsfont = ATSFontFindFromPostScriptName(CFSTR("Times-Italic"), 
    kATSOptionFlagsDefault);
    else
    atsfont = ATSFontFindFromPostScriptName(CFSTR("Times-Roman"), 
    kATSOptionFlagsDefault);
    CGFontRef cgfont = CGFontCreateWithPlatformFont(&atsfont);

    CGContextSetFont(context, cgfont);
    CGContextSetFontSize(context,fontPtr->fa.size);

    int symbol[512];
    for (int ii=0; ii<512; ii++)
    symbol[ii] = ii-29;

    int ll = strlen(text);
    CGGlyph gg[256];
    for (int ii=0; ii<ll; ii++)
    gg[ii] = symbol[text[ii]];

    CGContextShowGlyphsAtPoint(context, vv[0], vv[1], gg, ll);
  */

  // done
  CGContextSynchronize(context);
  QDEndCGContext(destPort, &context);
}

extern "C" {
  void XDrawString(Display* display, Drawable drawable, GC gc, 
			  int x, int y, _Xconst char* str, int length)
  {
    XDrawRotString(display, drawable, gc, (Tk_Font)gc->font, 
    		   0, Vector(x,y), str);
    //Tk_DrawChars(display, drawable, gc, (Tk_Font)gc->font, str, length, x, y);
  }

  // Pixmap Suite
  Pixmap XCreatePixmap(Display* display, Drawable drawable, 
			      unsigned int width, unsigned int height,
			      unsigned int depth)
  {
    Tk_GetPixmap(display, drawable, width, height, depth);
  }

  // XImage Suite
  XImage* XXCreateImage(Display* display, Visual* visual,
			unsigned int depth, int format, int offset,
			char* data, unsigned int width, unsigned int height,
			int bitmap_pad, int bytes_per_line)
  {
    XImage* ximage = XCreateImage(display, visual, depth, format, offset,
				  data, width, height, 
				  bitmap_pad, bytes_per_line);

    GWorldPtr gw = (GWorldPtr)data;
    PixMapHandle ph = GetGWorldPixMap(gw);
    PixMap* pm = *ph;

    // now correct it
    ximage->bytes_per_line = (pm->rowBytes & 0x3fff);

    switch (depth) {
    case 1:
      ximage->bits_per_pixel = 1;
      ximage->bitmap_unit = 8;
      ximage->bitmap_pad = 1;
      break;
    case 8:
      ximage->bits_per_pixel = 8;
      ximage->bitmap_unit = 8;
      ximage->bitmap_pad = 8;
      break;
    case 15:
      ximage->bits_per_pixel = 16;
      ximage->bitmap_unit = 32;
      ximage->bitmap_pad = 32;
      break;
    case 16:
      ximage->bits_per_pixel = 16;
      ximage->bitmap_unit = 32;
      ximage->bitmap_pad = 32;
      break;
    case 24:
      ximage->bits_per_pixel = 24;
      ximage->bitmap_unit = 32;
      ximage->bitmap_pad = 32;
      break;
    case 32:
      ximage->bits_per_pixel = 32;
      ximage->bitmap_unit = 32;
      ximage->bitmap_pad = 32;
      break;
    }

    switch (pm->pixelFormat) {
    case k16BE555PixelFormat: // 0x10
      ximage->bitmap_bit_order = MSBFirst;
      ximage->byte_order = MSBFirst;

      ximage->red_mask   = 0x00007C00;
      ximage->green_mask = 0x000003E0;
      ximage->blue_mask  = 0x0000001F;
      break;
    case k16LE555PixelFormat: // 'L555'
      ximage->bitmap_bit_order = LSBFirst;
      ximage->byte_order = LSBFirst;

      ximage->red_mask   = 0x00007C00;
      ximage->green_mask = 0x000003E0;
      ximage->blue_mask  = 0x0000001F;
      break;
    case k16BE565PixelFormat: // 'B565'
      ximage->bitmap_bit_order = MSBFirst;
      ximage->byte_order = MSBFirst;

      ximage->red_mask   = 0x0000F800;
      ximage->green_mask = 0x000007E0;
      ximage->blue_mask  = 0x0000001F;
      break;
    case k16LE565PixelFormat: // 'L565'
      ximage->bitmap_bit_order = LSBFirst;
      ximage->byte_order = LSBFirst;

      ximage->red_mask   = 0x0000F800;
      ximage->green_mask = 0x000007E0;
      ximage->blue_mask  = 0x0000001F;
      break;

    case k24RGBPixelFormat: // 0x18
      ximage->bitmap_bit_order = MSBFirst;
      ximage->byte_order = MSBFirst;

      ximage->red_mask   = 0x00FF0000;
      ximage->green_mask = 0x0000FF00;
      ximage->blue_mask  = 0x000000FF;
      break;
    case k24BGRPixelFormat: // '24BG'
      ximage->bitmap_bit_order = LSBFirst;
      ximage->byte_order = LSBFirst;

      ximage->red_mask   = 0x00FF0000;
      ximage->green_mask = 0x0000FF00;
      ximage->blue_mask  = 0x000000FF;
      break;

    case k32ARGBPixelFormat: // 0x20
      ximage->bitmap_bit_order = MSBFirst;
      ximage->byte_order = MSBFirst;

      ximage->red_mask   = 0x00FF0000;
      ximage->green_mask = 0x0000FF00;
      ximage->blue_mask  = 0x000000FF;
      break;
    case k32BGRAPixelFormat: // 'BGRA'
      ximage->bitmap_bit_order = LSBFirst;
      ximage->byte_order = LSBFirst;

      ximage->red_mask   = 0x00FF0000;
      ximage->green_mask = 0x0000FF00;
      ximage->blue_mask  = 0x000000FF;
      break;
    case k32ABGRPixelFormat: // 'ABGR'
      ximage->bitmap_bit_order = MSBFirst;
      ximage->byte_order = MSBFirst;

      ximage->red_mask   = 0x000000FF;
      ximage->green_mask = 0x0000FF00;
      ximage->blue_mask  = 0x00FF0000;
      break;
    case k32RGBAPixelFormat: // 'RGBA'
      ximage->bitmap_bit_order = LSBFirst;
      ximage->byte_order = LSBFirst;

      ximage->red_mask   = 0x000000FF;
      ximage->green_mask = 0x0000FF00;
      ximage->blue_mask  = 0x00FF0000;
      break;

    default:
      cerr << "MacOSX Internal Error: XXCreateImage: pixelFormat " << hex << pm->pixelFormat 
	   << " not implemented" << endl;
      break;
    }

    return ximage;
  }

  XImage *XXGetImage(Display *display, Drawable drawable, 
		     int x, int y,
		     unsigned int width, unsigned int height,
		     unsigned long plane_mask, int format)
  { 
    if (width ==0 || height ==0) {
      cerr << "MacOSX Internal Error: XXGetImage: width/height is zero" << endl;
      return NULL;
    }

    if (format != ZPixmap) {
      cerr << "MacOSX Internal Error: XXGetImage: illegal format" << endl;
      return NULL;
    }

    if (!TkMacOSXGetDrawablePort(drawable)) {
      cerr << "MacOSX Internal Error: XXGetImage: bad drawable" << endl;
      return NULL;
    }

    // get pixmap
    Pixmap pixmap = Tk_GetPixmap(display, drawable, width, height, 32);
    if (!pixmap) {
      cerr << "MacOSX Internal Error: XXGetImage: bad pixmap" << endl;
      return NULL;
    }
 
    // get gc
    GC gc;
    Tk_Window win = (Tk_Window)((MacDrawable *) drawable)->winPtr;
    if (win) {
      XGCValues values;
      gc = Tk_GetGC(win, 0, &values);
    } 
    else
      gc = XCreateGC(display, pixmap, 0, NULL);
 
    // populate pixmap
    XCopyArea(display, drawable, pixmap, gc, x, y, width, height, 0, 0);

    if (win)
      Tk_FreeGC(display, gc);
    else
      XFreeGC(display, gc);

    GWorldPtr gw = TkMacOSXGetDrawablePort(pixmap);
    PixMapHandle ph = GetGWorldPixMap(gw);
    PixMap* pm = *ph;

    int depth = pm->pixelSize;
    int bitmap_pad = depth; // will be reset
    int bytes_per_line = width; // will be reset
    XImage* image = XXCreateImage(display, NULL, depth, format, 0,
				     (char*)TkMacOSXGetDrawablePort(pixmap),
				     width, height, 
				     bitmap_pad, bytes_per_line);
    // needed by tk
    image->obdata = (XPointer)pixmap;

    return image;
  }

  void XSetClipRectangles(Display* display, GC gc, int clip_x_origin,
			  int  clip_y_origin, XRectangle* r,
			  int  n, int  ordering)
  {
    // do we already have a rgn?
    TkRegion clipRgn = NULL;
    if (gc->clip_mask) {
      TkpClipMask* msk = (TkpClipMask*)gc->clip_mask;
      if (msk->type == TKP_CLIP_REGION) {
	SetEmptyRgn((RgnHandle)msk->value.region);
	clipRgn = msk->value.region;
      }
    }

    // we may need a new region
    if (!clipRgn)
      clipRgn = TkCreateRegion();

    gc->clip_x_origin = clip_x_origin;
    gc->clip_y_origin = clip_y_origin;
    for (int ii =0; ii<n; ii++)
      TkUnionRectWithRegion(&r[ii], clipRgn, clipRgn);

    TkSetRegion(display, gc, clipRgn);
  }

  Bool XXQueryPointer(Display* display, Window w, 
		      Window* root_return, Window* child_return, 
		      int* root_x_return, int* root_y_return, 
		      int* win_x_return, int* win_y_return, 
		      unsigned int* mask_return)
  {
    XQueryPointer(display, w, root_return, child_return, 
		  root_x_return, root_y_return, 
		  win_x_return, win_y_return, mask_return);

    // flag root for XXWarpPointer
    *root_return = 1;
    *child_return = NULL;
  }

  void XXWarpPointer(Display* display, Window src_w, Window dest_w,
		     int src_x, int src_y,
		     unsigned int src_width, unsigned int src_height,
		     int dest_x, int dest_y)
  {
    // we have a special mode, root=1, from XXQueryPointer
    CGPoint pt;

    if (dest_w == None) {
      Point ppt;
      GetGlobalMouse(&ppt);
      pt.x = dest_x+ppt.h;
      pt.y = dest_y+ppt.v;
    }
    else if (dest_w == 1) {
      pt.x = dest_x;
      pt.y = dest_y;
    }
    else {
      Point ppt;
      MacDrawable* macPtr = (MacDrawable*)dest_w;
      ppt.h = dest_x+macPtr->xOff;
      ppt.v = dest_y+macPtr->yOff;

      CGrafPtr port = TkMacOSXGetDrawablePort(dest_w);
      QDLocalToGlobalPoint(port, &ppt);
      pt.x = ppt.h;
      pt.y = ppt.v;
    }

    // we can't use the following, we want a mouse event generated.
    //      CGWarpMouseCursorPosition(pt);
    // and we can't use the following, its buggy as all get out
    //      CGPostMouseEvent(pt, true, 1, false);

    CGEventSourceRef src = CGEventSourceCreate(kCGEventSourceStateCombinedSessionState);
    CGEventRef ev = CGEventCreateMouseEvent(src,kCGEventMouseMoved,pt,kCGMouseButtonLeft);
    // bug in tiger, reset event type
    CGEventSetType(ev,kCGEventMouseMoved);
    CGEventPost(kCGSessionEventTap,ev);
    CFRelease(ev);
    CFRelease(src);
  }
}
